


from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import FaktsGrant, GrantException
from fakts.beacon import EndpointDiscovery, FaktsRetriever
from rich.prompt import Prompt
from rich.console import Console


class PrompingBeaconGrantException(GrantException):
    pass

class NoBeaconsFound(PrompingBeaconGrantException):
    pass


class CLIBeaconGrant(FaktsGrant):

    def __init__(self, *args, dicovery_protocol: EndpointDiscovery = None, retriever_protocol: FaktsRetriever = None, timeout=4, console = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self._discov = dicovery_protocol or EndpointDiscovery()
        self._retriev = retriever_protocol or FaktsRetriever()
        self._console = console or Console()
        self._timeout = timeout

    async def aload(self, previous = {}, **kwargs):

        with self._console.status(f"Waiting {self._timeout} seconds for Beacon Answers"):
            endpoints = await self._discov.ascan_list(timeout=self._timeout)
        
        if len(endpoints.keys()) == 0: raise NoBeaconsFound("We couldn't find any beacon in your local network")

        choices_name = [key for key, value in endpoints.items()]
        endpoint_name = Prompt.ask("Which Endpoint do you want", choices=choices_name, default=choices_name[0])

        with self._console.status(f"Please check your broswer window to finish the setup"):
            return await self._retriev.aretrieve(endpoints[endpoint_name], previous)




from fakts.grants.base import FaktsGrant
import yaml


class YamlGrant(FaktsGrant):

    def __init__(self, filepath = "random.yaml") -> None:
        super().__init__()
        self.filepath = filepath


    async def aload(self, previous = {}, **kwargs):
        with open(self.filepath,"r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        return config





from fakts.grants.base import GrantException, FaktsGrant
import yaml
from koil.qt import FutureWrapper
from qtpy import QtWidgets


class NoFileSelected(GrantException):
    pass



class QtYamlGrant(FaktsGrant, QtWidgets.QWidget):

    def __init__(self) -> None:
        super().__init__()
        self.frequest = FutureWrapper()
        self.frequest.call.connect(self.on_open_filepath)


    def on_open_filepath(self, ref):
        self.show()
        filepath, weird =  QtWidgets.QFileDialog.getOpenFileName(self, 'Hey! Select a File')
        if filepath:
            self.frequest.resolve.emit(ref, filepath)
        else:
            self.frequest.reject.emit(ref, NoFileSelected("User did not select a File"))

        self.hide()


    async def aload(self, previous = {}, **kwargs):
        filepath = await self.frequest.acall()
        with open(filepath,"r") as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
        
        return config



from typing import List
from fakts.middleware.base import FaktsMiddleware
from fakts.utils import update_nested
from koil import koil
import yaml
from fakts.grants.base import FaktsGrant
import os
from fakts.grants.yaml import YamlGrant
from fakts.middleware.environment.overwritten import OverwrittenEnvMiddleware
import logging
import sys

logger = logging.getLogger(__name__)


class Fakts:
    def __init__(
        self,
        *args,
        grants=[YamlGrant(filepath="bergen.yaml")],
        middlewares=[OverwrittenEnvMiddleware()],
        fakts_path="fakts.yaml",
        register=True,
        force_reload=False,
        subapp: str = None,
        hard_fakts={},
        **kwargs,
    ) -> None:
        super().__init__(*args, **kwargs)
        self.loaded = False
        self.grants: List[FaktsGrant] = grants
        self.middlewares: List[FaktsMiddleware] = middlewares
        self.grantExceptions = []
        self.fakts = hard_fakts
        self.failedResponses = {}
        self.subapp = subapp
        self.fakts_path = f"{subapp}.{fakts_path}" if subapp else fakts_path

        try:
            config = self.load_config_from_file()
            self.fakts = update_nested(self.fakts, config)
            self.loaded = True
            logger.info(
                f"Loaded fakts from local file {self.fakts_path}. Delete this file or pass force_reload to Fakts"
            )
        except:
            logger.info(
                f"Couldn't load local conf-file {self.fakts_path}. We will have to refetch!"
            )

        if register:
            set_current_fakts(self)

    def load_config_from_file(self, filepath=None):
        with open(filepath or self.fakts_path, "r") as file:
            return yaml.load(file, Loader=yaml.FullLoader)

    async def aget(self, group_name, bypass_middleware=False):
        assert self.loaded, "Konfik needs to be loaded before we can access call load()"
        config = {**self.fakts}

        if not bypass_middleware:
            for middleware in self.middlewares:
                additional_config = await middleware.aparse(previous=config)
                config = update_nested(config, additional_config)

        for subgroup in group_name.split("."):
            try:
                config = config[subgroup]
            except KeyError as e:
                logger.error(f"Could't find {subgroup} in {config}")
                config = {}

        return config

    async def arefresh(self):
        await self.aload()

    async def aload(self):
        assert (
            len(self.grants) > 0
        ), "Please provide allowed Grants to retrieve the Fakts from"

        for grant in self.grants:
            try:
                additional_fakts = await grant.aload(previous=self.fakts)
                self.fakts = update_nested(self.fakts, additional_fakts)
                self.grantExceptions.append(None)
            except Exception as e:
                self.grantExceptions.append(e)

        assert (
            self.fakts != {}
        ), f"We did not received any valid Responses from our Grants"

        if self.fakts_path:
            with open(self.fakts_path, "w") as file:
                yaml.dump(self.fakts, file)

        self.loaded = True

    async def adelete(self):
        self.loaded = False
        self.fakts = None

        if self.fakts_path:
            os.remove(self.fakts_path)

    def load(self, **kwargs):
        return koil(self.aload(), **kwargs)

    def delete(self, **kwargs):
        return koil(self.adelete(), **kwargs)


CURRENT_FAKTS = None


def get_current_fakts(**kwargs) -> Fakts:
    global CURRENT_FAKTS
    if not CURRENT_FAKTS:
        CURRENT_FAKTS = Fakts(**kwargs)
    return CURRENT_FAKTS


def set_current_fakts(fakts) -> Fakts:
    global CURRENT_FAKTS
    if CURRENT_FAKTS:
        logger.error(
            "Hmm there was another fakts set, maybe thats cool but more likely not"
        )
    CURRENT_FAKTS = fakts



from qtpy.QtCore import Signal
from fakts.grants.beacon import BeaconGrant
from koil.koil import get_current_koil
from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import GrantException, FaktsGrant
from fakts.beacon import EndpointDiscovery, FaktsRetriever
from qtpy import QtWidgets
import asyncio
import logging


logger = logging.getLogger(__name__)


class RetrieveDialog(QtWidgets.QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.button = QtWidgets.QPushButton("Cancel")
        self.layout = QtWidgets.QHBoxLayout()
        self.layout.addWidget(self.button)
        self.setLayout(self.layout)


class SelfScanWidget(QtWidgets.QWidget):
    user_endpoint = Signal(FaktsEndpoint)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QHBoxLayout()
        self.lineEdit = QtWidgets.QLineEdit()
        self.addButton = QtWidgets.QPushButton("Scan")

        self.layout.addWidget(self.lineEdit)
        self.layout.addWidget(self.addButton)
        self.addButton.clicked.connect(self.on_add)
        self.setLayout(self.layout)

    def on_add(self):
        host = self.lineEdit.text()
        url = f"http://{host}:3000/setupapp"
        endpoint = FaktsEndpoint(url=url, name="Self Added")
        self.user_endpoint.emit(endpoint)


class UserCancelledException(GrantException):
    pass


class QtSelectableBeaconGrant(BeaconGrant, QtWidgets.QDialog):
    new_endpoint = Signal(FaktsEndpoint)
    show_signal = Signal()
    hide_signal = Signal()

    retrieve_start = Signal()
    retrieve_end = Signal()

    def __init__(self, *args, timeout=6000, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.retrieve_dialog = RetrieveDialog()
        self.retrieve_dialog.button.clicked.connect(self.on_cancel_retrieval)

        self.timeout = timeout
        self.new_endpoint.connect(self.on_new_endpoint)
        self.show_signal.connect(self.on_show)
        self.hide_signal.connect(self.on_hide)

        self.retrieve_start.connect(self.show_retrieve_dialog)
        self.retrieve_end.connect(self.hide_retrieve_dialog)

        self.endpoints = []
        self.layout = QtWidgets.QVBoxLayout()
        self.listWidget = QtWidgets.QListWidget()

        self.scanWidget = SelfScanWidget()
        self.cancelButton = QtWidgets.QPushButton("Cancel")
        self.scanWidget.user_endpoint.connect(self.on_endpoint_selected)
        self.cancelButton.clicked.connect(self.on_cancel)

        self.select_endpoint = None
        self.retrieving_task = None

        self.layout.addWidget(self.listWidget)
        self.layout.addWidget(self.scanWidget)
        self.layout.addWidget(self.cancelButton)
        self.setLayout(self.layout)

    def on_endpoint_clicked(self, item):
        index = self.listWidget.indexFromItem(item).row()
        return self.on_endpoint_selected(self.endpoints[index])

    def on_endpoint_selected(self, endpoint):
        self.loop.call_soon_threadsafe(self.select_endpoint.set_result, endpoint)

    def on_cancel(self, endpoint):
        self.loop.call_soon_threadsafe(
            self.select_endpoint.set_exception,
            UserCancelledException("User cancelled the Selection"),
        )

    def on_cancel_retrieval(self, endpoint):
        self.loop.call_soon_threadsafe(
            self.retrieving_task.set_exception,
            UserCancelledException("User cancelled the Retrieval"),
        )

    def on_show(self):
        self.show()

    def on_hide(self):
        self.hide()

    def show_retrieve_dialog(self):
        self.retrieve_dialog.show()

    def hide_retrieve_dialog(self):
        self.retrieve_dialog.hide()

    def closeEvent(self, event):
        # do stuff
        if self.select_endpoint:
            self.loop.call_soon_threadsafe(
                self.select_endpoint.set_exception,
                UserCancelledException("User cancelled the Selection"),
            )
        if self.retrieving_task:
            self.loop.call_soon_threadsafe(
                self.select_endpoint.set_exception,
                UserCancelledException("User cancelled the Selection"),
            )

        event.accept()  # let the window close

    def on_new_endpoint(self, config):
        self.listWidget.clear()

        self.endpoints.append(config)

        for endpoint in self.endpoints:
            self.listWidget.addItem(f"{endpoint.name} at {endpoint.url}")

        self.listWidget.itemClicked.connect(self.on_endpoint_clicked)

    async def emit_endpoints(self):

        async for endpoint in self._discov.ascan_gen():
            self.new_endpoint.emit(endpoint)

    async def aload(self, previous={}, **kwargs):
        self.loop = asyncio.get_event_loop()
        self.show_signal.emit()
        emitting_task = self.loop.create_task(self.emit_endpoints())
        self.select_endpoint = self.loop.create_future()

        konfik = None
        try:
            endpoint = await self.select_endpoint

            self.retrieve_start.emit()
            self.retrieving_task = self.loop.create_task(
                self._retriev.aretrieve(endpoint, previous=previous)
            )

            konfik = await asyncio.wait_for(self.retrieving_task, timeout=self.timeout)

            self.retrieving_task.cancel()

            try:
                await self.retrieving_task
            except asyncio.CancelledError as e:
                pass

            self.retrieve_end.emit()
            self.hide_signal.emit()

            return konfik

        finally:
            emitting_task.cancel()
            try:
                await emitting_task
            except asyncio.CancelledError as e:
                logger.info("Cancelled the Discovery task")



from fakts import Fakts
from qtpy import QtCore, QtWidgets
from koil.qt import FutureWrapper


class QtFakts(Fakts, QtWidgets.QWidget):
    loaded_signal = QtCore.Signal(bool)
    error_signal = QtCore.Signal(Exception)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.layout = QtWidgets.QVBoxLayout()
        self.setWindowTitle("Retrieval Wizard")

        self.showf = FutureWrapper()
        self.showf.wire(self.handle_show)
        self.showfref = None

        self.hidef = FutureWrapper()
        self.hidef.wire(self.handle_hide)

        self.title = QtWidgets.QLabel("Konfig Wizard")
        self.title.setAlignment(QtCore.Qt.AlignCenter)
        self.layout.addWidget(self.title)
        self.subtitle = QtWidgets.QLabel("These are the grants we are able to use")
        self.layout.addWidget(self.subtitle)

        for grant in self.grants:
            grant_title = QtWidgets.QLabel(grant.__class__.__name__)
            self.layout.addWidget(grant_title)

        self.start_button = QtWidgets.QPushButton("Start")
        self.start_button.clicked.connect(self.on_start)
        self.layout.addWidget(self.start_button)

        self.setLayout(self.layout)

    def handle_show(self, ref, *args, **kwargs):
        self.show()
        self.showfref = ref

    def handle_hide(self, ref, *args, **kwargs):
        self.close()
        self.hidef.resolve(ref, None)

    def on_start(self):
        self.showf.resolve(self.showfref, None)

    async def aload(self):
        await self.showf.acall()  # await user starts
        try:
            nana = await super().aload()
            self.loaded_signal.emit(True)
            await self.hidef.acall()
            return nana
        except Exception as e:
            self.error_signal.emit(e)
            await self.hidef.acall()
            raise e

    async def adelete(self):
        nana = await super().adelete()
        self.loaded_signal.emit(False)
        return nana






from fakts.beacon.beacon import FaktsEndpoint
from fakts.grants.base import FaktsGrant
from fakts.beacon import EndpointDiscovery, FaktsRetriever




class EndpointGrant(FaktsGrant):

    def __init__(self, endpoint: FaktsEndpoint, retrieve_protocol: FaktsRetriever = None) -> None:
        self._endpoint = endpoint
        self._retriev = retrieve_protocol  or FaktsRetriever()

        super().__init__()

    async def aload(self, previous={}, **kwargs):
        return await self._retriev.aretrieve(self._endpoint, previous=previous)